"""SniTun worker for traffics."""
import asyncio
from multiprocessing import Process, Manager, Queue
from threading import Thread
from typing import Dict, Optional, List
from socket import socket

from .listener_peer import PeerListener
from .listener_sni import SNIProxy
from .peer_manager import PeerManager, PeerManagerEvent
from .peer import Peer


class ServerWorker(Process):
    """Worker for multiplexer."""

    def __init__(
        self,
        fernet_keys: List[str],
        throttling: Optional[int] = None,
    ) -> None:
        """Initialize worker & communication."""
        super().__init__()

        self._fernet_keys: List[str] = fernet_keys
        self._throttling: Optional[int] = throttling

        # Used on the child
        self._peers: Optional[PeerManager] = None
        self._list_sni: Optional[SNIProxy] = None
        self._list_peer: Optional[PeerListener] = None

        # Communication between Parent/Child
        self._manager: Manager = Manager()
        self._new: Queue = self._manager.Queue()
        self._sync: Dict[str, None] = self._manager.dict()

    @property
    def peer_size(self) -> int:
        """Return amount of managed peers."""
        return len(self._sync)

    def is_responsible_peer(self, sni: str) -> bool:
        """Return True if worker is responsible for this peer domain."""
        return sni in self._sync

    async def _async_init(self) -> None:
        """Initialize child process data."""
        self._peers = PeerManager(
            self._fernet_keys,
            throttling=self._throttling,
            event_callback=self._event_stream,
        )
        self._list_sni = SNIProxy(self._peers)
        self._list_peer = PeerListener(self._peers)

    def _event_stream(self, peer: Peer, event: PeerManagerEvent) -> None:
        """Event stream peer connection data."""
        if event == PeerManagerEvent.CONNECTED:
            self._sync[peer.hostname] = None
        else:
            self._sync.pop(peer.hostname, None)

    def shutdown(self) -> None:
        """Shutdown child process."""
        self._new.put(None)
        self.join(10)

    def handover_connection(
        self, con: socket, data: bytes, sni: Optional[str] = None
    ) -> None:
        """Move new connection to worker."""
        self._new.put_nowait((con, data, sni))

    def run(self) -> None:
        """Running worker process."""
        loop = asyncio.new_event_loop()

        # Start eventloop
        running_loop = Thread(target=loop.run_forever)
        running_loop.start()

        # Init backend
        asyncio.run_coroutine_threadsafe(self._async_init(), loop=loop).result()

        while True:
            new = self._new.get()
            if new is None:
                break

            new[0].setblocking(False)
            loop.call_soon_threadsafe(
                asyncio.create_task, self._async_new_connection(*new)
            )

        # Shutdown worker
        loop.call_soon_threadsafe(loop.stop)
        running_loop.join(10)

    async def _async_new_connection(
        self, con: socket, data: bytes, sni: Optional[str]
    ) -> None:
        """Handle incoming connection."""
        try:
            reader, writer = await asyncio.open_connection(sock=con)
        except OSError:
            con.close()
            return

        # Select the correct handler for process connection
        loop = asyncio.get_running_loop()
        if sni:
            loop.create_task(
                self._list_sni.handle_connection(reader, writer, data=data, sni=sni)
            )
        else:
            loop.create_task(
                self._list_peer.handle_connection(reader, writer, data=data)
            )



"""Represent a single Peer."""
import asyncio
from datetime import datetime
import hashlib
import logging
import os
from typing import Optional, Coroutine

from ..exceptions import MultiplexerTransportDecrypt, SniTunChallengeError
from ..multiplexer.core import Multiplexer
from ..multiplexer.crypto import CryptoTransport

_LOGGER = logging.getLogger(__name__)


class Peer:
    """Representation of a Peer."""

    def __init__(
        self,
        hostname: str,
        valid: datetime,
        aes_key: bytes,
        aes_iv: bytes,
        throttling: Optional[int] = None,
    ):
        """Initialize a Peer."""
        self._hostname = hostname
        self._valid = valid
        self._throttling = throttling
        self._multiplexer = None
        self._crypto = CryptoTransport(aes_key, aes_iv)

    @property
    def hostname(self) -> str:
        """Return his hostname."""
        return self._hostname

    @property
    def is_connected(self) -> bool:
        """Return True if we are connected to peer."""
        if not self._multiplexer:
            return False
        return self._multiplexer.is_connected

    @property
    def is_valid(self) -> bool:
        """Return True if the peer is valid."""
        return self._valid > datetime.utcnow()

    @property
    def multiplexer(self) -> Optional[Multiplexer]:
        """Return Multiplexer object."""
        return self._multiplexer

    @property
    def is_ready(self) -> bool:
        """Return true if the Peer is ready to process data."""
        if self.multiplexer is None:
            return False
        if not self.multiplexer.is_connected:
            return False
        return True

    async def init_multiplexer_challenge(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Initialize multiplexer."""
        try:
            token = hashlib.sha256(os.urandom(40)).digest()
            writer.write(self._crypto.encrypt(token))
            await writer.drain()

            data = await reader.readexactly(32)
            data = self._crypto.decrypt(data)

            # Check Token
            assert hashlib.sha256(token).digest() == data

        except (
            asyncio.IncompleteReadError,
            MultiplexerTransportDecrypt,
            AssertionError,
            OSError,
        ) as err:
            _LOGGER.warning("Wrong challenge from peer")
            raise SniTunChallengeError() from err

        # Start Multiplexer
        self._multiplexer = Multiplexer(
            self._crypto, reader, writer, throttling=self._throttling
        )

    def wait_disconnect(self) -> Coroutine:
        """Wait until peer is disconnected.

        Return a coroutine.
        """
        if not self._multiplexer:
            raise RuntimeError("No Transport initialize for peer")

        return self._multiplexer.wait()
